import numpy as np
import sklearn.linear_model as scireg

def euclidean_distance(point1, point2):
    return np.linalg.norm(point1 - point2)

class Calibrator:

    def __init__(self):
        self.X = []
        self.Y_y = []
        self.Y_x = []
        self.reg = None
        self.reg_x = scireg.Ridge(alpha=1.0)
        self.reg_y = scireg.Ridge(alpha=1.0)
        self.fitted = False

    def add(self,x,y):
        self.X.append(x.flatten())
        self.Y_y.append(y[1])
        self.Y_x.append(y[0])

        __tmp_X =np.array(self.X)
        __tmp_Y_y =np.array(self.Y_y)
        __tmp_Y_x =np.array(self.Y_x)

        self.reg_x.fit(__tmp_X,__tmp_Y_x)
        self.reg_y.fit(__tmp_X,__tmp_Y_y)
        self.fitted = True

    def predict(self,x):
        if self.fitted:
            x = x.flatten()
            x = x.reshape(1, -1)
            y_x = self.reg_x.predict(x)[0]
            y_y = self.reg_y.predict(x)[0]
            return np.array([y_x,y_y])
        else:
            return np.array([0.0,0.0])

    def unfit(self):
        self.fitted = False

class CalibrationMatrix:

    def __init__(self):

        self.iterator = 0
        self.points = np.array([[1,0.5],[0.75,0.5],[0.5,0.5],[0.25,0.5],[0.0,0.5],
                                [1.0,1.0],[0.75,1.0],[0.5,1.0],[0.25,1.0],[0.0,1.0],
                                [1.0,0.0],[0.75,0.0],[0.5,0.0],[0.25,0.0],[0.0,0.0],
                                [0.0,0.75],[0.75,0.75],[0.5,0.75],[0.25,0.75],[0.0,0.75],
                                [1.0,0.25],[0.75,0.25],[0.5,0.25],[0.25,0.25],[0.0,0.25]])
        pass

    def getNextPoint(self,width=1.0,height=1.0):
        it = self.iterator
        self.iterator += 1
        self.iterator %= len(self.points)

        return np.array([self.points[it,0] * width, self.points[it,1] * height])


import pytest
import eyeGestures.screenTracker.screenTracker as scrtr
import eyeGestures.screenTracker.dataPoints as dp

# region of interest inside the virtual screen
ROI_WIDTH = 30
ROI_HEIGHT = 20

####################### screen 2 display Tests #################################


def test_scaleUpTest():
    pos = (50, 50)
    BIGGER = 10

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    roi2 = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH+BIGGER, ROI_HEIGHT+BIGGER)

    bigger_roi = scrtr.scaleUp(roi, roi2, 0.1)

    assert bigger_roi.width > roi.width
    assert bigger_roi.height > roi.height


def test_scaleUpTest_known():
    pos = (50, 50)
    BIGGER = 10

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    roi2 = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH+BIGGER, ROI_HEIGHT+BIGGER)

    bigger_roi = scrtr.scaleUp(roi, roi2, 0.1)

    assert int(bigger_roi.width) == int(ROI_WIDTH*1.1)
    assert int(bigger_roi.height) == int(ROI_HEIGHT*1.1)


def test_scaleDownTest():
    pos = (50, 50)
    SMALLER = 10

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    roi2 = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH-SMALLER, ROI_HEIGHT-SMALLER)

    smaller_roi = scrtr.scaleDown(roi, roi2, -0.1)

    assert smaller_roi.width < roi.width
    assert smaller_roi.height < roi.height


def test_scaleUpTest_known():
    pos = (50, 50)
    SMALLER = 10

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    roi2 = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH-SMALLER, ROI_HEIGHT-SMALLER)

    smaller_roi = scrtr.scaleDown(roi, roi2, -0.1)

    assert int(smaller_roi.width) == int(ROI_WIDTH*0.9)
    assert int(smaller_roi.height) == int(ROI_HEIGHT*0.9)



"""Module providing a extraction of eye from face object."""

import cv2
import numpy as np
import mediapipe as mp


class Eye:
    """Class storing data related and representing a eye"""
    LEFT_EYE_KEYPOINTS = np.array(
        list(mp.solutions.face_mesh.FACEMESH_LEFT_EYE))[:, 0]
    RIGHT_EYE_KEYPOINTS = np.array(
        list(mp.solutions.face_mesh.FACEMESH_RIGHT_EYE))[:, 0]
    LEFT_EYE_IRIS_KEYPOINT = []
    RIGHT_EYE_IRIS_KEYPOINT = []
    LEFT_EYE_PUPIL_KEYPOINT = [473]
    RIGHT_EYE_PUPIL_KEYPOINT = [468]

    # LEFT_EYE_KEYPOINTS = [36, 37, 38, 39, 40, 41] # keypoint indices for left eye
    # RIGHT_EYE_KEYPOINTS = [42, 43, 44, 45, 46, 47] # keypoint indices for right eye

    scale = (150, 100)

    def __init__(self, side: int):

        # check if eye is left or right
        if side == 1:
            self.side = "right"
            self.pupil_index = self.RIGHT_EYE_PUPIL_KEYPOINT
        elif side == 0:
            self.side = "left"
            self.pupil_index = self.LEFT_EYE_PUPIL_KEYPOINT

        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.center_x = 0
        self.center_y = 0
        self.image = None
        self.pupil = None
        self.offset = None
        self.region = None
        self.cut_image = None
        self.landmarks = None

        # self._process(self.image,self.region)

    def update(self, image: np.ndarray, landmarks: list, offset: np.ndarray):
        """function updating data stored inside eye object"""

        self.image = image
        self.offset = offset
        self.landmarks = landmarks

        # check if eye is left or right
        if self.side == "right":
            self.region = np.array(
                landmarks[self.RIGHT_EYE_KEYPOINTS], dtype=np.int32)

        elif self.side == "left":
            self.region = np.array(
                landmarks[self.LEFT_EYE_KEYPOINTS], dtype=np.int32)

        self.pupil = landmarks[self.pupil_index][0]
        self._process(self.image, self.region)

    def getCenter(self):
        """function returning center of eye"""

        return (self.center_x, self.center_y)

    def getPos(self):
        """function returning position of eye in the image"""

        return (self.x, self.y)

    def getPupil(self):
        """function returning pupil object"""

        # return self.pupil.getCoords()
        return self.pupil

    def getBlink(self):
        """function returning blink event"""
        # return self.pupil.getCoords()
        return (self.height) <= 3  # 2x margin

    def getImage(self):
        """function returning image of the eye cut from the entire face image"""

        # TODO: draw additional parameters
        return self.cut_image

    def getGaze(self, gaze_buffor, y_correction=0, x_correction=0):
        """function returning gaze position"""

        # pupilCoords = self.pupil.getCoords()
        center = np.array((self.center_x, self.center_y)) - self.offset

        region_corrected = self.region - self.offset
        pupil_corrected = self.pupil - self.offset

        vectors = region_corrected - center
        pupil = pupil_corrected - center

        vectors = vectors - pupil
        gaze_vector = np.zeros((2))

        gaze_vector[1] = np.sum(vectors, axis=0)[1] * 10 - y_correction
        gaze_vector[0] = -np.sum(vectors, axis=0)[0] * 10 - x_correction

        # print("gaze_vector: ",gaze_vector)
        gaze_buffor.add(gaze_vector)
        return gaze_buffor.getAvg()

    def getOpenness(self):
        """function returning eye openness"""

        return self.height/2

    def getLandmarks(self):
        """function returning eye landmarks"""

        return self.region
    
    def getBoundingBox(self):
        return (self.x,self.y,self.width,self.height)

    def _process(self, image, region):
        h, w, _ = image.shape

        mask = np.full((h, w), 0, dtype=np.uint8)
        background = np.zeros((h, w), dtype=np.uint8)
        cv2.fillPoly(mask, [region], 0)

        masked_image = cv2.bitwise_not(background, cv2.cvtColor(
            image.copy(), cv2.COLOR_BGR2GRAY), mask=mask)

        margin = 2
        min_x = np.min(region[:, 0]) - margin
        max_x = np.max(region[:, 0]) + margin
        min_y = np.min(region[:, 1]) - margin
        max_y = np.max(region[:, 1]) + margin

        self.x = min_x
        self.y = min_y

        self.width = np.max(region[:, 0]) - np.min(region[:, 0])
        self.height = np.max(region[:, 1]) - np.min(region[:, 1])

        self.center_x = (min_x + max_x)/2
        self.center_y = (min_y + max_y)/2

        # HACKETY_HACK:
        self.pupil[1] = np.min(region[:, 1])

        self.cut_image = masked_image[min_y:max_y, min_x:max_x]
        # print(f"here: {self.cut_image.shape,min_y,max_y,min_x,max_x}")
        self.cut_image = cv2.cvtColor(self.cut_image, cv2.COLOR_GRAY2BGR)

        for point in self.region:
            point = point - (min_x, min_y)
            cv2.circle(self.cut_image, point.astype(
                int), 1, (255, 0, 0, 150), 1)

        pupil = self.pupil - (min_x, min_y)

        # print(f"pupil: {pupil}")
        cv2.circle(self.cut_image, pupil.astype(int), 1, (0, 255, 0, 150), 1)

        self.cut_image = cv2.resize(self.cut_image, self.scale)

        # save cut_image to buffor and get avg from previous buffors
        # self.eyeBuffer.add(self.cut_image)
        # self.cut_image = np.array(self.eyeBuffer.getAvg(), dtype=np.uint8)

        # LEGACY
        # org_scale = (max_x - min_x,max_y - min_y)
        # self.pupil = pupil.Pupil(self.cut_image, min_x, min_y, self.scale, org_scale)



import time
import queue
import pickle
import platform
import threading

import cv2
import numpy as np

# Make predictions for new data points


def timeit(func):
    """
    timeit
    """
    def inner(*args, **kwargs):
        """
        inner
        """
        start = time.time()
        ret = func(*args, **kwargs)
        print(f"Elapsed time: {time.time() - start}")
        return ret
    return inner


def shape_to_np(shape, dtype="int"):
    """
    shape_to_np
    """
    coords = np.zeros((68, 2), dtype=dtype)
    for i in range(0, 68):
        coords[i] = (shape.part(i).x, shape.part(i).y)
    return coords


def make_image_grid(images, rows, cols):
    """
    Make a grid of images.

    Parameters:
    - images: list of images to form the grid (all images must be of the same size and type)
    - rows, cols: number of rows and columns in the grid

    Returns:
    - grid_image: image grid as a single image
    """
    # Check if the list of images is not empty and that we have enough to fill the grid
    assert images, "List of images is empty"
    # assert len(images) >= rows * cols, "Not enough images to fill the grid"

    # Get image dimensions
    img_h, img_w = images[0].shape[:2]

    if len(images[0].shape) > 2:
        # Create a black canvas to draw the grid on
        grid_image = np.zeros(
            (img_h * rows, img_w * cols, images[0].shape[2]), dtype=np.uint8)
    else:
        grid_image = np.zeros((img_h * rows, img_w * cols), dtype=np.uint8)

    # Copy images to the grid
    for i, img in enumerate(images):
        if i >= rows * cols:
            break  # Stop if we have filled the grid
        row = i // cols
        col = i % cols
        grid_image[row * img_h:(row + 1) * img_h, col *
                   img_w:(col + 1) * img_w] = img

    return grid_image


class var:

    def __init__(self, var):
        self.__var = var

    def set(self, var):
        self.__var = var

    def get(self):
        return self.__var


class Buffor:

    def __init__(self, length):
        self.length = length
        self.__buffor = []

    def add(self, var):
        if len(self.__buffor) >= self.length:
            self.__buffor.pop(0)

        self.__buffor.append(var)

    def getAvg(self, lenght=0):
        return np.sum(self.__buffor[-lenght:], axis=0) / len(self.__buffor[-lenght:])

    def getBuffor(self):
        return self.__buffor

    def loadBuffor(self, buffor):
        self.__buffor = buffor

    def getLast(self):
        return self.__buffor[0]

    def getFirst(self):
        return self.__buffor[len(self.__buffor) - 1]

    def getLen(self):
        return len(self.__buffor)
    
    def flush(self):
        tmp = self.__buffor[-1]
        self.__buffor = []
        self.__buffor.append(tmp)

    def clear(self):
        self.__buffor = []

# Bufforless


class VideoCapture:
    """Wrapper on openCV2 stream making it bufforless and adding camera search"""

    def __init__(self, name, bufforless=True):
        self.bufforless = bufforless
        self.run = True

        if isinstance(name, str):
            if ".pkl" in name:
                self.stream = False
            else:
                self.stream = True
        else:
            self.stream = True

        if self.stream:
            self.prev_frame = None

            self.__openCam(name)

            self.q = queue.Queue()
            self.t = threading.Thread(target=self.__reader)
            self.t.start()
        else:
            self.frames = []
            with open(name, 'rb') as file:
                self.frames = pickle.load(file)

    def __openCam(self, name):
        if isinstance(name, int):
            if "Windows" in platform.system():
                self.cap = cv2.VideoCapture(name, cv2.CAP_DSHOW)
            else:
                self.cap = cv2.VideoCapture(name)

            if self.cap is None or not self.cap.isOpened():
                print(f"Was unable to open camera: {name}.")
                print(f"Trying to open camera: {name}.")
                if name + 1 < 10:
                    self.__openCam(name + 1)
        else:
            self.cap = cv2.VideoCapture(name)

    def __reader(self):
        while self.run:
            ret, frame = self.cap.read()
            if not ret:
                break
            if not self.q.empty() and self.bufforless:
                try:
                    self.q.get_nowait()
                except queue.Empty:
                    pass
            self.q.put((ret, frame))

        self.flush()

    def flush(self):
        while not self.q.empty():
            self.q.get()

    def read(self):
        """Function returning latest frame"""
        if self.stream:
            return self.q.get()
        else:
            frame = self.frames.pop(0)
            self.frames.pop(0)
            return ((len(self.frames) >= 1), frame)

    def close(self):
        """Function closing stream"""
        self.run = False
        self.t.join()
        self.cap.release()



import numpy as np

class Heatmap():
    """Helper representing Heatmap of tracked points"""

    def __init__(self,width,height,buffor):
        self.inc_step = 10
        self.step = 10
        bars = self.step

        self.width = width
        self.height = height

        bars_x = int(width/self.step)
        bars_y = int(height/self.step)

        self.axis_x = np.zeros((bars_x))
        self.axis_y = np.zeros((bars_y))

        for point in buffor:
            x = point[0]
            y = point[1]

            self.axis_x[min(abs(int(x/bars)),bars_x - 1)] += self.inc_step
            self.axis_y[min(abs(int(y/bars)),bars_x - 1)] += self.inc_step

        self.min_x = self.__getParam((self.axis_x > self.inc_step*4),last=False)
        self.max_x = self.__getParam((self.axis_x > self.inc_step*4),last=True)
        self.min_y = self.__getParam((self.axis_y > self.inc_step*4),last=False)
        self.max_y = self.__getParam((self.axis_y > self.inc_step*4),last=True)

    def __getParam(self,param,last : bool = False):
        ret = 0
        retArray = np.where(param)

        if len(retArray[0]) > 0:
            ret = retArray[0][- int(last)] * self.step

            if not ret == np.NAN:
                ret = int(ret)

        return ret

    def getBoundaries(self):
        """Function returning boundaries of heatmap"""
        x = self.min_x
        y = self.min_y
        w = self.max_x - self.min_x
        h = self.max_y - self.min_y
        return (x,y,w,h)

    def getCenter(self):
        """Function returning center of heatmap"""
        center_x = int((self.max_x - self.min_x)/2 + self.min_x)
        center_y = int((self.max_y - self.min_y)/2 + self.min_y)
        return (center_x,center_y)

    def getPeak(self):
        """Function returning peak of heatmap"""
        x = int(np.argmax(self.axis_x)*self.inc_step)
        y = int(np.argmax(self.axis_y)*self.inc_step)
        return (x,y)

    def getHist(self):
        """Function returning histogram of the heatmap. This heatmap is simply 2d histogram, so this function returns values for each axis."""
        return (self.axis_x/self.inc_step,self.axis_y/self.inc_step)



"""Module providing a core tracking features."""

import eyeGestures.screenTracker.dataPoints as dp
from eyeGestures.Fixation import Fixation
from eyeGestures.utils import Buffor


class Contexter:
    """Class storing contextes"""

    def __init__(self):
        self.context = dict()

    def addContext(self, context_id, object):
        """Function allowing to add new context to contexter"""

        if context_id not in self.context.keys():
            self.context[context_id] = object
            return True
        return False

    def rmContext(self, context_id):
        """Function allowing to removes context from contexter"""

        if context_id in self.context.keys():
            del self.context[context_id]
            return True
        return False

    def getContext(self, context_id):
        """Function returning context based on id"""

        if context_id in self.context.keys():
            return self.context[context_id]
        return None

    def updateContext(self, context_id, data):
        """Function updating context with new data"""

        if context_id in self.context.keys():
            self.context[context_id] = data
            return True
        self.addContext(context_id, data)
        return True

    def getNumberContextes(self):
        """Function returning number of contextes"""

        return len(self.context.keys())


class Gcontext:
    """Helper class for Gcontext"""

    def __init__(self,
                 display,
                 face,
                 roi,
                 edges,
                 cluster_boundaries,
                 gazeBuffor,
                 l_pupil,
                 r_pupil,
                 l_eye_buff,
                 r_eye_buff,
                 fixation,
                 calibration):

        self.roi = roi
        self.face = face
        self.edges = edges
        self.cluster_boundaries = cluster_boundaries
        self.gazeBuffor = gazeBuffor
        self.l_pupil = l_pupil
        self.r_pupil = r_pupil
        self.l_eye_buff = l_eye_buff
        self.r_eye_buff = r_eye_buff
        self.display = display
        self.fixation = fixation
        self.calibration = calibration


class GazeContext:
    """Context wrapper for gaze tracker application"""

    def __init__(self):
        self.contexter = Contexter()

    def get(self,
            id,
            display,
            face=None,
            roi=dp.ScreenROI(285, 105, 80, 15),
            edges=dp.ScreenROI(285, 105, 80, 15),
            cluster_boundaries=dp.ScreenROI(225, 125, 20, 20),
            buffor=Buffor(200),
            l_pupil=Buffor(20),
            r_pupil=Buffor(20),
            l_eye_buff=Buffor(20),
            r_eye_buff=Buffor(20),
            fixation=Fixation(0, 0, 100),
            calibration=False):
        """Function creating new context or returning if id already exists"""

        context = Gcontext(display=display,
                           face=face,
                           roi=roi,
                           edges=edges,
                           cluster_boundaries=cluster_boundaries,
                           gazeBuffor=buffor,
                           l_pupil=l_pupil,
                           r_pupil=r_pupil,
                           l_eye_buff=l_eye_buff,
                           r_eye_buff=r_eye_buff,
                           fixation=fixation,
                           calibration=calibration)

        if self.contexter.addContext(id, context):
            return context
        else:
            return self.contexter.getContext(id)

    def update(self,
               id,
               context):
        """Function updating existing context"""

        self.contexter.updateContext(id, context)



import math
import numpy as np

from scipy import signal
from sklearn.cluster import DBSCAN

import eyeGestures.screenTracker.dataPoints as dp
from eyeGestures.screenTracker.clusters import Clusters
from eyeGestures.screenTracker.heatmap import Heatmap

# THIS FILE IS SLOWLY BECOMING BLACK MAGIC

def detect_if_inside(point,rect):
    px = point[0]
    py = point[1]
    x,y,width,height = rect.getBoundaries() 

    x_in = x < px and px < x + width
    y_in = y < py and py < y + height
    
    return x_in and y_in

def detect_edges(roi, display, point_on_screen, point_on_display):
    """Function performing edge detection based on point, screen and display sizes"""
    (s_x,s_y) = point_on_screen 
    (d_x,d_y) = point_on_display

    x,y,width,height = roi.getBoundaries() 
    new_roi = dp.ScreenROI(x,y,width,height)

    # new_w = screen.width
    # margin = 5
    # TODO: fix this to estimate w and h
    if d_x <= 0 :
        new_roi.width = new_roi.width + abs(new_roi.x - s_x)
        new_roi.x = s_x

    if d_x >= display.width :
        new_roi.width = abs(new_roi.x - s_x)

    if d_y <= 0 :
        new_roi.height = new_roi.height + abs(new_roi.y - s_y)
        new_roi.y = s_y

    if d_y >= display.height :
        new_roi.height = abs(new_roi.y - s_y)

    return new_roi

def rescale_h(roi, scale_h, change = 0.5):
    """change size on height"""
    scale_diff_h = abs(1.0 - scale_h)

    ret_heigth  = roi.height

    if scale_diff_h > change: 
        ret_heigth = roi.height/1.0 * scale_h

    return ret_heigth

def rescale_w(roi, scale_w, change = 0.5):
    """change size on width"""
    scale_diff_w = abs(1.0 - scale_w)

    ret_width   = roi.width

    if scale_diff_w > change: 
        ret_width = roi.width/1.0 * scale_w

    return ret_width


def scaleDown(roi, edge, scale):
    """Function scalling down one roi till edges are met"""

    (_,_,cluster_w,cluster_h) = edge.getBoundaries()

    new_roi = dp.ScreenROI(roi.x,roi.y,roi.width,roi.height)
    if cluster_w < roi.width :
        new_scale_w = 1.0 + scale
        new_w = rescale_w(roi,new_scale_w,scale)
        new_roi.width = new_w

    if cluster_h < roi.height :
        new_scale_h = 1.0 + scale
        new_h = rescale_h(roi,new_scale_h,scale)
        new_roi.height = new_h

    return new_roi

def scaleUp(roi, roi2, scale):
    """Function scalling up one roi into another roi, second roi is for limiting scaling operation and scale is telling how much you can scale"""

    (_,_,roi2_w,roi2_h) = roi2.getBoundaries()

    new_roi = dp.ScreenROI(roi.x,roi.y,roi.width,roi.height)

    if roi2_w > roi.width :
        new_scale_w = 1.0 + scale
        new_w = rescale_w(roi,new_scale_w,scale)
        new_roi.width = new_w

    if roi2_h > roi.height :
        new_scale_h = 1.0 + scale
        new_h = rescale_h(roi,new_scale_h,scale)
        new_roi.height = new_h

    return new_roi


class ScreenProcessor:
    """Class for doing tracking and converting screen to display and display to screen"""

    def __init__(self):
        pass

    # RUN FEATURE
    def process(self,point,point_offset,buffor_length,roi,edges,screen,display,heatmap):
        """Function performing main processing for screen processor. Tons of parameters doing brrrr"""

        s_point_offset = self.display2screen(point_offset,screen,display)
        p_on_display = self.screen2display([point[0] + s_point_offset[0],point[1] + s_point_offset[1]],roi,display)
        # p_on_display = self.screen2display(point,roi,display)

        if buffor_length > 20:
            new_edges = detect_edges(roi, display, point, p_on_display)
            edges.x      = new_edges.x
            edges.y      = new_edges.y
            edges.width  = new_edges.width
            edges.height = new_edges.height

        p_on_display = (p_on_display[0] + display.offset_x, p_on_display[1] + display.offset_y)

        # return how close that is hist region
        (_,_,roi_w,roi_h) = heatmap.getBoundaries()
        closeness_percentage = (roi_w * roi_h)/(screen.width * screen.height)
        return (p_on_display,closeness_percentage)

    # CALIBRATION FEATURE
    # it should not affect objects but produce new ones
    def update(self, roi, edges, cluster, heatmap):
        """Function to update screen processor with new clusters and heatmaps"""

        (x,y) = heatmap.getCenter()
        # =====================================
        # ---------histogram obtained----------
        # =====================================
        new_roi = dp.ScreenROI(roi.x,roi.y,roi.width,roi.height)

        new_roi.setCenter(x,y)
        edges.setCenter(x,y)
        # self.eyeScreen.setCenter(x,y) 
        new_roi = scaleUp(new_roi, edges, scale = 0.1)

        # =====================================
        # if screen is bigger than edges make it smaller
        # =====================================
        new_roi = scaleDown(new_roi, cluster, scale = -0.1)
        return new_roi


    def screen2display(self, screen_point, screen, display):
        """Function converting screen points to display points"""

        s_x,s_y = screen_point[0],screen_point[1]

        d_x = int((s_x - screen.x)/screen.width * display.width)
        d_y = int((s_y - screen.y)/screen.height * display.height)

        d_x = max(d_x,0)
        d_y = max(d_y,0)
        d_x = min(d_x,display.width)
        d_y = min(d_y,display.height)

        return (d_x,d_y)

    def display2screen(self, display_point, screen, display):
        """Function converting display points to screen points"""

        d_x,d_y = display_point[0],display_point[1]

        s_x = int((d_x)/display.width  * screen.width)
        s_y = int((d_y)/display.height * screen.height)


        return (s_x,s_y)

class ScreenManager:
    """Class performing most of tracking after estimating position of gaze"""

    def __init__(self,):
        self.screen_processor = ScreenProcessor()


    def process(self, buffor, roi, edges, screen, display, calibration, offset):
        """Function doing processing and tracking and calibration of tracker"""

        heatmap = Heatmap(screen.width,screen.height,buffor.getBuffor())
        cluster = Clusters(buffor.getBuffor()).getMainCluster()

        if cluster is not None:


            if calibration:
                roi = self.screen_processor.update(roi, edges, cluster, heatmap)

            p, percentage = self.screen_processor.process(
                buffor.getAvg(20),
                (offset[0],offset[1]),
                len(buffor.getBuffor()),
                roi,
                edges,
                screen,
                display,
                heatmap)

            return (p, roi, cluster)
        return ([0,0], roi, cluster)




import numpy as np
from eyeGestures.gevent import Gevent
from eyeGestures.face import FaceFinder, Face
from eyeGestures.Fixation import Fixation
from eyeGestures.processing import EyeProcessor
from eyeGestures.gazeContexter import GazeContext
from eyeGestures.screenTracker.screenTracker import ScreenManager
import eyeGestures.screenTracker.dataPoints as dp
from eyeGestures.utils import Buffor


def isInside(circle_x, circle_y, r, x, y):
    """Function checking if points are inside the circle"""

    # Compare radius of circle
    # with distance of its center
    # from given point
    if ((x - circle_x) * (x - circle_x) +
            (y - circle_y) * (y - circle_y) <= r * r):
        return True
    else:
        return False


class GazeTracker:
    """Class processing images and returning gaze tracker"""

    N_FEATURES = 16

    def __init__(self, screen_width, screen_heigth,
                 eye_screen_w, eye_screen_h,
                 roi_x, roi_y,
                 roi_width, roi_height,
                 monitor_offset_x=0,
                 monitor_offset_y=0):

        self.screen = dp.Screen(screen_width, screen_heigth)

        self.offset_x = 0
        self.offset_y = 0

        self.roi_x = roi_x
        self.roi_y = roi_y
        self.roi_width = roi_width
        self.roi_height = roi_height

        self.eye_screen_w = eye_screen_w
        self.eye_screen_h = eye_screen_h

        self.eyeProcessorLeft = EyeProcessor(eye_screen_w, eye_screen_h)
        self.eyeProcessorRight = EyeProcessor(eye_screen_w, eye_screen_h)

        self.screen_man = ScreenManager()

        self.finder = FaceFinder()

        # those are used for analysis
        self.__headDir = [0.5, 0.5]

        self.point_screen = [0.0, 0.0]
        self.freezed_point = [0.0, 0.0]

        self.face = Face()
        self.GContext = GazeContext()
    #     self.calibration = False

    def __gaze_intersection(self, l_eye, r_eye, l_buff, r_buff):
        l_pupil = l_eye.getPupil()
        l_gaze = l_eye.getGaze(l_buff)

        r_pupil = r_eye.getPupil()
        r_gaze = r_eye.getGaze(r_buff)

        l_end = l_gaze + l_pupil
        r_end = r_gaze + r_pupil

        l_m = (l_end[1] - l_pupil[1])/(l_end[0] - l_pupil[0])
        r_m = (r_end[1] - r_pupil[1])/(r_end[0] - r_pupil[0])

        l_b = l_end[1] - l_m * l_end[0]
        r_b = r_end[1] - r_m * r_end[0]

        i_x = (r_b - l_b)/(l_m - r_m)
        i_y = r_m * i_x + r_b
        return (i_x, i_y)

    def __pupil(self, eye, eyeProcessor, intersection_x, buffor):

        eyeProcessor.append(eye.getPupil(), eye.getLandmarks(), buffor)
        point = eyeProcessor.getAvgPupil(
            self.eye_screen_w, self.eye_screen_h, buffor)
        point = np.array((int(intersection_x), point[1]))

        return point, buffor

    def estimate(self,
                 image,
                 display,
                 context_id,
                 calibration,
                 fixation_freeze=0.7,
                 freeze_radius=20,
                 offset_x=0,
                 offset_y=0):
        """Function estimating gaze and returning gaze event based on image"""

        event = None
        face_mesh = self.getFeatures(image)
        if not face_mesh:
            return None

        self.face.process(image, face_mesh)

        context = self.GContext.get(context_id, display,
            face = None,
            roi =dp.ScreenROI(
                self.roi_x,
                self.roi_y,
                self.roi_width,
                self.roi_height),
            edges=dp.ScreenROI(285, 105, 80, 15),
            cluster_boundaries=dp.ScreenROI(225, 125, 20, 20),
            buffor=Buffor(200),
            l_pupil=Buffor(20),
            r_pupil=Buffor(20),
            l_eye_buff=Buffor(20),
            r_eye_buff=Buffor(20),
            fixation=Fixation(0, 0, 100))
        context.calibration = calibration

        if not self.face is None:

            if context.face == None:
                x,y,w,h=self.face.getBoundingBox()
                i_w = self.face.image_w
                i_h = self.face.image_h
                context.face = (x,y,w,h,i_w,i_h)


            l_eye = self.face.getLeftEye()
            r_eye = self.face.getRightEye()

            # TODO: check what happens here before with l_pupil
            intersection_x, _ = self.__gaze_intersection(
                l_eye, r_eye, context.l_eye_buff, context.r_eye_buff)
            l_point, l_buffor = self.__pupil(
                l_eye, self.eyeProcessorLeft,  intersection_x, context.l_pupil)
            r_point, r_buffor = self.__pupil(
                r_eye, self.eyeProcessorRight, intersection_x, context.r_pupil)

            context.l_pupil = l_buffor
            context.r_pupil = r_buffor

            compound_point = np.array(((l_point + r_point)/2), dtype=np.uint32)

            blink = l_eye.getBlink() or r_eye.getBlink()
            if blink != True:
                context.gazeBuffor.add(compound_point)

            if blink != True:
                # current face radius
                face_x,face_y,face_w,face_h = self.face.getBoundingBox()
                image_w = self.face.image_w
                image_h = self.face.image_h

                face_w_perc = face_w/image_w
                face_h_perc = face_h/image_h

                # # prev current face radius
                c_face_x,c_face_y,c_face_w,c_face_h,c_image_w,c_image_h = context.face

                c_face_w_perc = c_face_w/c_image_w
                c_face_h_perc = c_face_h/c_image_h

                x,y,w,h=self.face.getBoundingBox()
                i_w = self.face.image_w
                i_h = self.face.image_h
                context.face = (x,y,w,h,i_w,i_h)

                # roi update

                # context.roi.x -= diff_face_x * 500 # current size of virtual display
                # context.roi.y -= diff_face_y * 500 # current size of virtual display
                if abs(face_w_perc/c_face_w_perc - 1.0) > 0.02:
                    context.roi.width  = context.roi.width * abs(face_w_perc/c_face_w_perc)
                    # context.edges.width= context.edges.width * abs(face_w_perc/c_face_w_perc)
                    context.gazeBuffor.flush()
                    # context.calibration = True
                if abs(face_h_perc/c_face_h_perc - 1.0) > 0.02:
                    context.roi.height = context.roi.height* abs(face_h_perc/c_face_h_perc)
                    # context.edges.height= context.edges.height * abs(face_w_perc/c_face_w_perc)
                    context.gazeBuffor.flush()
                    # context.calibration = True

            self.point_screen, roi, cluster = self.screen_man.process(context.gazeBuffor,
                                                                      context.roi,
                                                                      context.edges,
                                                                      self.screen,
                                                                      context.display,
                                                                      context.calibration,
                                                                      (offset_x,
                                                                       offset_y)
                                                                      )

            context.roi = roi
            if cluster:
                x, y, width, height = cluster.getBoundaries()
                context.cluster_boundaries.x = x
                context.cluster_boundaries.y = y
                context.cluster_boundaries.width = width
                context.cluster_boundaries.height = height

            self.GContext.update(context_id, context)

            ###########################################################

            fix = context.fixation.process(
                self.point_screen[0], self.point_screen[1])
            # this should prevent of sudden movement down when blinking - not perfect yet

            if fix > fixation_freeze:
                r = freeze_radius
                if not isInside(self.freezed_point[0], self.freezed_point[1], r, self.point_screen[0], self.point_screen[1]):
                    self.freezed_point = self.point_screen

                event = Gevent(self.freezed_point,
                               blink,
                               fix,
                               l_eye,
                               r_eye,
                               display,
                               context.roi,
                               context.edges,
                               context.cluster_boundaries,
                               context_id)
            else:
                self.freezed_point = self.point_screen
                event = Gevent(self.point_screen,
                               blink,
                               fix,
                               l_eye,
                               r_eye,
                               display,
                               context.roi,
                               context.edges,
                               context.cluster_boundaries,
                               context_id)

        return event

    # def get_contextes(self):
    #     """Function returning contextes"""

    #     return self.GContext.get_contextes()

    # def add_offset(self,x,y):
    #     """Function adding offsets to tracker"""

    #     self.screen_man.push_window(x,y)

    def getFeatures(self, image):
        """Function returning face landmarks"""

        face_mesh = self.finder.find(image)
        return face_mesh




from eyeGestures.gazeEstimator import GazeTracker
import eyeGestures.screenTracker.dataPoints as dp
from eyeGestures.calibration_v1 import Calibrator as Calibrator_v1
from eyeGestures.calibration_v2 import Calibrator as Calibrator_v2, CalibrationMatrix, euclidean_distance
from eyeGestures.gevent import Gevent, Cevent
from eyeGestures.utils import timeit
import numpy as np
import cv2

import random

VERSION = "2.0.0"

class EyeGestures_v2:
    """Main class for EyeGesture tracker. It configures and manages entire algorithm"""


    PRECISION_LIMIT = 50
    PRECISION_STEP = 10
    ACCEPTANCE_RADIUS = 500
    CALIBRATION_RADIUS = 1000
    EYEGESTURES_CALIBRATION_THRESH = 850

    def __init__(self):

        self.monitor_width  = 1
        self.monitor_height = 1

        self.clb = Calibrator_v2()
        self.cap = None
        self.gestures = EyeGestures_v1(285,115,40,15)

        self.calibration = False

        self.CN = 5

        self.trackerSignal = None
        self.fitSignal = None

        self.average_points = np.zeros((20,2))
        self.filled_points = 0
        self.calibrate_gestures = True
        self.calibrationMat = CalibrationMatrix()
        self.fit_point = self.calibrationMat.getNextPoint()

        self.iterator = 0
        self.fix = 0.8

        self.precision_limit = self.PRECISION_LIMIT
        self.precision_step = self.PRECISION_STEP
        self.acceptance_radius = self.ACCEPTANCE_RADIUS
        self.calibration_radius = self.CALIBRATION_RADIUS
        # after corssing this thresh we are disabling classical calib
        self.eyegestures_calibration_threshold = self.EYEGESTURES_CALIBRATION_THRESH

    def getLandmarks(self, frame, calibrate = False):

        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        frame = cv2.flip(frame,1)
        # frame = cv2.resize(frame, (360, 640))

        event, _ = self.gestures.step(
            frame,
            "main",
            calibrate, # set calibration - switch to False to stop calibration
            self.monitor_width,
            self.monitor_height,
            0, 0, self.fix, 100)

        cursor_x, cursor_y = event.point[0],event.point[1]
        l_eye_landmarks = event.l_eye.getLandmarks()
        r_eye_landmarks = event.r_eye.getLandmarks()

        cursors = np.array([cursor_x,cursor_y]).reshape(1, 2)
        eye_events = np.array([event.blink,event.fixation]).reshape(1, 2)
        key_points = np.concatenate((cursors,l_eye_landmarks,r_eye_landmarks,eye_events))
        return  np.array((cursor_x, cursor_y)), key_points, event.blink, event.fixation

    def increase_precision(self):
        if self.acceptance_radius > self.precision_limit:
            self.acceptance_radius -= self.precision_step
        if self.calibration_radius > self.precision_limit and self.acceptance_radius < self.calibration_radius:
            self.calibration_radius -= self.precision_step

    def setClassicImpact(self,impact):
        self.CN = impact

    def reset(self):
        self.acceptance_radius = self.ACCEPTANCE_RADIUS
        self.calibration_radius = self.CALIBRATION_RADIUS
        self.average_points = np.zeros((20,2))
        self.filled_points = 0
        self.clb.unfit()

    def setFixation(self,fix):
        self.fix = fix

    def setClassicalImpact(self,CN):
        self.CN = CN

    def step(self, frame, calibration, width, height):
        self.calibration = calibration
        self.monitor_width = width
        self.monitor_height = height

        classic_point, key_points, blink, fixation = self.getLandmarks(frame,self.calibrate_gestures)

        margin = 10
        if classic_point[0] <= margin and self.calibration:
            self.calibrate_gestures = True
        elif classic_point[0] >= width - margin and self.calibration:
            self.calibrate_gestures = True
        elif classic_point[1] <= margin and self.calibration:
            self.calibrate_gestures = True
        elif classic_point[1] >= height - margin and self.calibration:
            self.calibrate_gestures = True
        else:
            self.calibrate_gestures = False

        y_point = self.clb.predict(key_points)
        self.average_points[1:,:] = self.average_points[:(self.average_points.shape[0] - 1),:]
        if fixation <= self.fix:
            self.average_points[0,:] = y_point

        if self.filled_points < self.average_points.shape[0] and (y_point != np.array([0.0,0.0])).any():
            self.filled_points += 1

        averaged_point = (np.sum(self.average_points[:,:],axis=0) + (classic_point * self.CN))/(self.filled_points + self.CN)

        if self.calibration and (euclidean_distance(averaged_point,self.fit_point) < self.calibration_radius or self.filled_points < self.average_points.shape[0] * 10):
            self.clb.add(key_points,self.fit_point)

        if euclidean_distance(averaged_point,self.fit_point) < self.acceptance_radius:
            self.iterator += 1
            if self.iterator > 10:
                self.iterator = 0
                self.fit_point = self.calibrationMat.getNextPoint(width,height)
                self.increase_precision()

        gevent = Gevent(averaged_point,blink,fixation >= self.fix)
        cevent = Cevent(self.fit_point,self.acceptance_radius, self.calibration_radius)
        return (gevent, cevent)

class EyeGestures_v1:
    """Main class for EyeGesture tracker. It configures and manages entier algorithm"""

    def __init__(self,
                 roi_x=225,
                 roi_y=105,
                 roi_width=80,
                 roi_height=15):

        screen_width = 500
        screen_height = 500
        height = 250
        width = 250

        roi_x = roi_x % screen_width
        roi_y = roi_y % screen_height
        roi_width = roi_width % screen_width
        roi_height = roi_height % screen_height

        self.screen_width = screen_width
        self.screen_height = screen_height

        self.screen = dp.Screen(
            screen_width,
            screen_height)

        self.gaze = GazeTracker(screen_width,
                                screen_height,
                                width,
                                height,
                                roi_x,
                                roi_y,
                                roi_width,
                                roi_height)
        
        self.calibrators = dict()
        self.calibrate = False

    def getFeatures(self, image):
        """[NOT RECOMMENDED] Function allowing for extraction of gaze features from image"""

        return self.gaze.getFeatures(image)

    # @timeit
    # 0.011 - 0.015 s for execution
    def step(self, image,
                 context,
                 calibration,
                 display_width,
                 display_height,
                 display_offset_x=0,
                 display_offset_y=0,
                 fixation_freeze=0.7,
                 freeze_radius=20,
                 offset_x=0,
                 offset_y=0):
        """Function performing estimation and returning event object"""

        display = dp.Display(
            display_width,
            display_height,
            display_offset_x,
            display_offset_y
        )

        # allow for random recalibration shot
        event = self.gaze.estimate(image,
                                  display,
                                  context,
                                  calibration,
                                  fixation_freeze,
                                  freeze_radius,
                                  offset_x,
                                  offset_y)

        cursor_x,cursor_y = event.point[0],event.point[1]
        if context in self.calibrators:
            self.calibrate = self.calibrators[context].calibrate(cursor_x,cursor_y,event.fixation)
        else:
            self.calibrators[context] = Calibrator_v1(display_width,display_height,cursor_x,cursor_y)
            self.calibrate = self.calibrators[context].calibrate(cursor_x,cursor_y,event.fixation)

        cpoint = self.calibrators[context].get_current_point()
        calibration = self.calibrate and cpoint != (0,0)
        cevent = Cevent(cpoint,100, 100, calibration)

        return (event, cevent)


"""Module providing a fixation detection."""


class Fixation:
    """Class performing Fixation"""

    def __init__(self, x, y, radius=100):
        self.radius = radius
        self.fixation = 0.0
        self.x = x
        self.y = y
        pass

    def process(self, x, y):
        """Function processing x and y estimated points for fixation detection"""

        if (x - self.x)**2 + (y - self.y)**2 < self.radius**2:
            self.fixation = min(self.fixation + 0.02, 1.0)
        else:
            self.x = x
            self.y = y
            self.fixation = 0

        return self.fixation
